/*
  socket_class.cc - Interface for network programming

  (c) 1998 Martin Pitt (Piware)

  This file is distributed under the terms and conditions of the 
  "GNU Library General Public License" (see file COPYING.LIB)

  Please report bugs and comments to piware@piware.de
*/

#include "socket_class.h"

#include <sys/socket.h> // socket(), accept(), bind() etc.
#include <netinet/in.h>	// hton[s/l](), INADDR_ANY
#include <netdb.h>	// gethostby[name/addr] ()
#include <stdio.h>	// sprintf()
#include <sys/time.h>	// struct timeval, select()
#include <fcntl.h>	// fcntl() and constants
#include <unistd.h>	// read(), write()

Socket::Socket( Port _port ) 
	: port( _port ), state( CLOSED )
{
    socketDesc = socket( AF_INET, SOCK_STREAM, 0 );
    int on = 1;
    setsockopt( socketDesc, SOL_SOCKET, SO_REUSEADDR, &on, sizeof( on ) );
}

Socket::Socket( int _desc, Port _port ) 
	: socketDesc( _desc ), port( _port ), state( CONNECTED )
{}

Socket::~Socket()
{
    Close();
}

void
Socket::Close()
{
    if( !Closed() )
     {
	 shutdown( socketDesc, 2 );
	 port = 0;
	 state = CLOSED;
     }
}

void
Socket::SetBlocking( bool b ) const
{
    int flags = fcntl( socketDesc, F_GETFL );
    fcntl( socketDesc, F_SETFL, b ? ( flags & ~O_NONBLOCK ) : ( flags |	O_NONBLOCK ) );
}

bool
Socket::IsBlocking() const
{
    return fcntl( socketDesc, F_GETFL ) & O_NONBLOCK == 0;
}

void
Socket::Listen( unsigned maxClients )
{
    	// Bind socket to port ********
    struct sockaddr_in addr;

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = htons( port );

    if( bind( socketDesc, (struct sockaddr *) &addr, sizeof( addr ) ) < 0 )
	throw ExcSocketSys( "Socket::Listen( unsigned ) : bind" );
        
	// Set to listen mode ****************
    if( listen( socketDesc, maxClients ) == -1 )
	throw ExcSocketSys( "Socket::Listen( unsigned ) : listen" );
    
    state = LISTENING;
};

Socket*
Socket::Accept() const
{
    int peer;
    socklen_t size = sizeof( sockaddr_in );
    struct sockaddr_in peeraddr;
    
    if( ( peer = accept( socketDesc, (struct sockaddr *) &peeraddr, &size ) ) == -1 )
	return NULL;
    
    return new Socket( peer, port );
}

bool
Socket::Connect( unsigned long ip )
{
    struct sockaddr_in addr;

    ip = htonl( ip );
    memcpy( &addr.sin_addr, &ip, sizeof( ip ) );
    addr.sin_family = AF_INET;
    addr.sin_port = htons( port );

	// connect ***************************
    if( connect( socketDesc, (struct sockaddr *) &addr, sizeof( addr ) ) == -1 )
	return false;

    state = CONNECTED;
    return true;
}

bool
Socket::Connect( unsigned char dq1, unsigned char dq2, 
		 unsigned char dq3, unsigned char dq4 )
{
    return Connect( (unsigned long) dq1 << 24 +
		    (unsigned long) dq2 << 16 +
		    (unsigned long) dq3 << 8  +
		    dq4 );
}

bool
Socket::DataAvail() const
{
    bool dAvail;
    char d;

    bool blocking = IsBlocking();
    SetBlocking( false );
    dAvail = ( recv( socketDesc, &d, 1, MSG_PEEK ) == 1 );
    errno = 0;
    SetBlocking( blocking );
    return dAvail;
}

bool
Socket::WaitRead( unsigned timeOut ) const
{
    fd_set fs;
    struct timeval tm;

    FD_ZERO( &fs );
    FD_SET( socketDesc, &fs );
    
    tm.tv_sec = timeOut;
    tm.tv_usec = 0;

    return select( FD_SETSIZE, &fs, NULL, NULL, &tm );
}

string
Socket::DottedQuadIP( unsigned long ip ) 
{
    char s[17];
    sprintf( s, "%i.%i.%i.%i", int( ip>>24 ), int( (ip>>16) & 0xFF ),
	     int( (ip>>8) & 0xFF ), int( ip & 0xFF ) );
    return *( new string( s ) );
}

unsigned long
Socket::Resolve( const char* hostname )
{
    struct hostent* host;

    if( !( host = gethostbyname( hostname ) ) )
	throw ExcSocketHostName( "unsigned long Socket::Resolve( const char* )" );

    unsigned long ip;
    memcpy( &ip, host->h_addr_list[0], sizeof( ip ) );
    return ntohl( ip );
}

string
Socket::Resolve( unsigned long ip )
{
    struct hostent* host;
    ip = htonl( ip );
    if( !( host = gethostbyaddr( (const char*) &ip, sizeof( ip ), AF_INET ) ) )
	return "";
     else
	 return string( host->h_name );
}

unsigned
Socket::Read( void* buffer, unsigned size ) const
{
    return read( socketDesc, buffer, size );
}

string
Socket::ReadString() const
{
    unsigned len;
    read( socketDesc, &len, sizeof( len ) );
    if( len ) {
	char* c = new char[len+1];
	c[len] = 0;
	read( socketDesc, c, len );
	string str( c );
	delete c;
	return str;
     }
     else
	 return "";
}

unsigned long
Socket::ReadLong () const
{
    unsigned long x;
    read( socketDesc, &x, sizeof( x ) );
    return ntohl( x );
}

unsigned short
Socket::ReadShort() const
{
    unsigned short x;
    read( socketDesc, &x, sizeof( x ) );
    return ntohs( x );
}

float
Socket::ReadFloat() const
{
    float x;
    read( socketDesc, &x, sizeof( x ) );
    return x;
}

double
Socket::ReadDouble() const
{
    double x;
    read( socketDesc, &x, sizeof( x ) );
    return x;
}

char
Socket::ReadChar() const
{
    char x;
    read( socketDesc, &x, sizeof( x ) );
    return x;
}

void
Socket::Write( const void* buffer, unsigned size ) const
{
    write( socketDesc, buffer, size );
}

void
Socket::WriteString( const string& str ) const
{
    int len = str.length();
    write( socketDesc, &len, sizeof( len ) );
    if (len)
	write( socketDesc, str.c_str(), len );
}

void
Socket::WriteLong( unsigned long x ) const
{
    x = htonl( x );
    write( socketDesc, &x, sizeof( x ) );
}

void
Socket::WriteShort( unsigned short x ) const
{
    x = htons( x );
    write( socketDesc, &x, sizeof( x ) );
}

void
Socket::WriteFloat( float x ) const
{
    write( socketDesc, &x, sizeof( x ) );
}

void
Socket::WriteDouble( double x ) const
{
    write( socketDesc, &x, sizeof( x ) );
}

void
Socket::WriteChar( char x ) const
{
    write( socketDesc, &x, sizeof( x ) );
}

const Socket&
Socket::operator >>( string& s ) const
{
    unsigned len;
    read( socketDesc, &len, sizeof( len ) );
    if( len ) {
	char* c = new char[len+1];
	c[len] = 0;
	read( socketDesc, c, len );
	s = c;
	delete c;
    }
     else
	 s = "";
    return *this;
}
